#include "tmesh.h"
#include "triangle.h"
#include <float.h>
#include "framebuffer.h"

#include <fstream>
#include <iostream>

using namespace std;

void TMesh::Initialize() {

  verts = 0;
  vertsN = 0;
  connectivity = 0;
  trianglesN = 0;
  projVerts = 0;
  normals = 0;
  colors = 0;
  tcs = 0;
  texture = 0;

}

TMesh::TMesh() {

  Initialize();

}

void TMesh::SetAABB(AABB aabb) {

  vertsN = 8;
  trianglesN = 12;
  Allocate();

  int vi = 0;

  verts[vi] = V3(aabb.corners[0][0], aabb.corners[1][1],
    aabb.corners[1][2]);
  vi++;
  verts[vi] = V3(aabb.corners[0][0], aabb.corners[0][1],
    aabb.corners[1][2]);
  vi++;
  verts[vi] = V3(aabb.corners[1][0], aabb.corners[0][1],
    aabb.corners[1][2]);
  vi++;
  verts[vi] = V3(aabb.corners[1][0], aabb.corners[1][1],
    aabb.corners[1][2]);
  vi++;

  verts[vi] = V3(aabb.corners[0][0], aabb.corners[1][1],
    aabb.corners[0][2]);
  vi++;
  verts[vi] = V3(aabb.corners[0][0], aabb.corners[0][1],
    aabb.corners[0][2]);
  vi++;
  verts[vi] = V3(aabb.corners[1][0], aabb.corners[0][1],
    aabb.corners[0][2]);
  vi++;
  verts[vi] = V3(aabb.corners[1][0], aabb.corners[1][1],
    aabb.corners[0][2]);
  vi++;

  int ti = 0;
  connectivity[3*ti+0] = 0;
  connectivity[3*ti+1] = 1;
  connectivity[3*ti+2] = 2;
  ti++;
  connectivity[3*ti+0] = 2;
  connectivity[3*ti+1] = 3;
  connectivity[3*ti+2] = 0;
  ti++;
  connectivity[3*ti+0] = 3;
  connectivity[3*ti+1] = 2;
  connectivity[3*ti+2] = 6;
  ti++;
  connectivity[3*ti+0] = 6;
  connectivity[3*ti+1] = 7;
  connectivity[3*ti+2] = 3;
  ti++;
  connectivity[3*ti+0] = 7;
  connectivity[3*ti+1] = 6;
  connectivity[3*ti+2] = 5;
  ti++;
  connectivity[3*ti+0] = 5;
  connectivity[3*ti+1] = 4;
  connectivity[3*ti+2] = 7;
  ti++;
  connectivity[3*ti+0] = 0;
  connectivity[3*ti+1] = 4;
  connectivity[3*ti+2] = 5;
  ti++;
  connectivity[3*ti+0] = 0;
  connectivity[3*ti+1] = 1;
  connectivity[3*ti+2] = 5;
  ti++;
  connectivity[3*ti+0] = 0;
  connectivity[3*ti+1] = 3;
  connectivity[3*ti+2] = 7;
  ti++;
  connectivity[3*ti+0] = 7;
  connectivity[3*ti+1] = 4;
  connectivity[3*ti+2] = 0;
  ti++;
  connectivity[3*ti+0] = 1;
  connectivity[3*ti+1] = 5;
  connectivity[3*ti+2] = 6;
  ti++;
  connectivity[3*ti+0] = 6;
  connectivity[3*ti+1] = 2;
  connectivity[3*ti+2] = 1;
  ti++;

}

void TMesh::Allocate() {

  if (verts)
    delete []verts;
  verts = new V3[vertsN];

  if (projVerts)
    delete []projVerts;
  projVerts = new V3[vertsN];

  if (tcs)
    delete []tcs;
  tcs = new float[2*vertsN];

  if (connectivity)
    delete []connectivity;
  connectivity = new unsigned int[3*trianglesN];

}

void TMesh::RenderPoints(V3 colorv, PPC *ppc, 
  FrameBuffer *fb) {

  int psize = 5;
  unsigned int color = colorv.GetColor();
  Project(ppc);
  for (int vi = 0; vi < vertsN; vi++) {
    if (projVerts[vi][0] == FLT_MAX)
      continue;
    int u = (int) projVerts[vi][0];
    int v = (int) projVerts[vi][1];
    if (colors) {
      fb->DrawPoint(u, v, psize, colors[vi].GetColor());
    }
    else {
      fb->DrawPoint(u, v, psize, color);
    }
  }

}


void TMesh::Project(PPC *ppc) {

  for (int vi = 0; vi < vertsN; vi++) {
    if (!ppc->Project(verts[vi], projVerts[vi])) {
      projVerts[vi] = V3(FLT_MAX, FLT_MAX, FLT_MAX);
    }
  }

}

void TMesh::RenderWireFrame(V3 colorv, PPC *ppc, 
  FrameBuffer *fb) {

  int psize = 5;
  unsigned int color = colorv.GetColor();
  Project(ppc);
  for (int ti = 0; ti < trianglesN; ti++) {
    for (int ei = 0; ei < 3; ei++) {
      int vi0 = connectivity[3*ti+ei];
      int vi1 = connectivity[3*ti+((ei+1)%3)];
      if (projVerts[vi0][0] == FLT_MAX)
        continue;
      if (projVerts[vi1][0] == FLT_MAX)
        continue;
      if (colors) {
        fb->DrawSegment(projVerts[vi0], projVerts[vi1],
          colors[vi0], colors[vi1]);
      }
      else {
        fb->DrawSegment(projVerts[vi0], projVerts[vi1], 
          color);
      }
    }
  }

}

void TMesh::RenderFilled(V3 colorv, PPC *ppc, 
  FrameBuffer *fb, FrameBuffer *shadowMap, PPC *shadowMapPPC) {

  unsigned int color = colorv.GetColor();
  Project(ppc);
  // LightVertices();
  for (int ti = 0; ti < trianglesN; ti++) {
    Triangle currTri;
    bool clipped = false;
    for (int j = 0; j < 3; j++) {
      int vind = connectivity[3*ti+j];
      if (projVerts[vind][j] == FLT_MAX)
        clipped = true;
      currTri.verts[j] = projVerts[vind];
      if (colors) {
        currTri.colors[j] = colors[vind];
      }
      if (texture) {
        currTri.tcs[2*j+0] = tcs[vind*2+0];
        currTri.tcs[2*j+1] = tcs[vind*2+1];
      }
    }
    currTri.color = currTri.colors[0];
    if (clipped)
      continue;
    currTri.Rasterize(fb, ppc, texture, shadowMap, shadowMapPPC);
  }

}


void TMesh::Load(char *fname) {

  ifstream ifs(fname, ios::binary);
  if (ifs.fail()) {
    cerr << "INFO: cannot open file: " << fname << endl;
    return;
  }

  ifs.read((char*)&vertsN, sizeof(int));
  char yn;
  ifs.read(&yn, 1); // always xyz
  if (yn != 'y') {
    cerr << "INTERNAL ERROR: there should always be vertex xyz data" << endl;
    return;
  }
  if (verts)
    delete verts;
  verts = new V3[vertsN];
  if (projVerts)
    delete projVerts;
  projVerts = new V3[vertsN];

  ifs.read(&yn, 1); // colors 3 floats
  if (colors)
    delete colors;
  colors = 0;
  if (yn == 'y') {
    colors = new V3[vertsN];
  }

  ifs.read(&yn, 1); // normals 3 floats
  if (normals)
    delete normals;
  normals = 0;
  if (yn == 'y') {
    normals = new V3[vertsN];
  }

  ifs.read(&yn, 1); // texture coordinates 2 floats
  float *tcs = 0; // don't have texture coordinates for now
  if (tcs)
    delete tcs;
  tcs = 0;
  if (yn == 'y') {
    tcs = new float[vertsN*2];
  }

  ifs.read((char*)verts, vertsN*3*sizeof(float)); // load verts

  if (colors) {
    ifs.read((char*)colors, vertsN*3*sizeof(float)); // load colors
  }

  if (normals)
    ifs.read((char*)normals, vertsN*3*sizeof(float)); // load normals

  if (tcs)
    ifs.read((char*)tcs, vertsN*2*sizeof(float)); // load texture coordinates

  ifs.read((char*)&trianglesN, sizeof(int));
  if (connectivity)
    delete connectivity;
  connectivity = new unsigned int[trianglesN*3];
  ifs.read((char*)connectivity, trianglesN*3*sizeof(unsigned int)); // read tiangles

  ifs.close();

  cerr << "INFO: loaded " << vertsN << " verts, " << trianglesN << " tris from " << endl << "      " << fname << endl;
  cerr << "      xyz " << ((colors) ? "rgb " : "") << ((normals) ? "nxnynz " : "") << ((tcs) ? "tcstct " : "") << endl;

}

V3 TMesh::GetCenter() {

  V3 ret(0.0f, 0.0f, 0.0f);
  for (int vi = 0; vi < vertsN; vi++) {
    ret = ret + verts[vi];
  }
  ret = ret / (float) vertsN;
  return ret;

}

void TMesh::TranslateVerts(V3 transv) {

  for (int vi = 0; vi < vertsN; vi++) {
    verts[vi] = verts[vi] + transv;
  }

}

void TMesh::GetAABB(AABB *&aabb) {

  aabb = new AABB();
  for (int vi = 0; vi < vertsN; vi++) {
    aabb->AddPoint(verts[vi]);
  }

}

void TMesh::SetCenter(V3 newCenter) {

  V3 currCenter = GetCenter();
  TranslateVerts(newCenter-currCenter);

}

void TMesh::Scale(float aabbDiagonal) {

  AABB *aabb;
  GetAABB(aabb);
  float currD = (aabb->corners[0]-aabb->corners[1]).Length();
  float scalef = aabbDiagonal/currD;
  V3 currCenter = GetCenter();
  TranslateVerts(currCenter*-1.0f);
  for (int vi = 0; vi < vertsN; vi++) {
    verts[vi] = verts[vi]*scalef;
  }
  TranslateVerts(currCenter);
}


void TMesh::Spin(V3 axis, float thetad) {

    V3 center = GetCenter();
    for (int vi = 0; vi < vertsN; vi++) {
      verts[vi] = verts[vi].RotatePointAboutArbitraryAxis(center, axis, thetad);
    }

}

void TMesh::SetRectangle(V3 center, V3 xaxis, V3 yaxis, float rw, float rh) {

  vertsN = 4;
  trianglesN = 2;
  Allocate();
  verts[0] = center - xaxis*rw/2.0f - yaxis*rh/2.0f;
  verts[1] = center + xaxis*rw/2.0f - yaxis*rh/2.0f;
  verts[2] = center + xaxis*rw/2.0f + yaxis*rh/2.0f;
  verts[3] = center - xaxis*rw/2.0f + yaxis*rh/2.0f;

  float tilefs = 4.0f*2.0f;
  float tileft = 3.0f*2.0f;
  tilefs = 1.0f;
  tileft = 1.0f;
  int ind = 0;
  tcs[ind*2+0] = 0.0f;
  tcs[ind*2+1] = tileft;
  ind++;
  tcs[ind*2+0] = tilefs;
  tcs[ind*2+1] = tileft;
  ind++;
  tcs[ind*2+0] = tilefs;
  tcs[ind*2+1] = 0.0f;
  ind++;
  tcs[ind*2+0] = 0.0f;
  tcs[ind*2+1] = 0.0f;
  ind++;

  connectivity[0] = 0;
  connectivity[1] = 1;
  connectivity[2] = 2;
  connectivity[3] = 2;
  connectivity[4] = 3;
  connectivity[5] = 0;

}

void TMesh::RenderHW() {

  glEnable(GL_DEPTH_TEST);

  if (false) {
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glLineWidth(3.0f);
  }
  else
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  glEnableClientState(GL_VERTEX_ARRAY);
  glVertexPointer(3, GL_FLOAT, 0, (float*)verts);

  if (colors) {
    glEnableClientState(GL_COLOR_ARRAY);
    glColorPointer(3, GL_FLOAT, 0, (float*)colors);
  }

  if (normals) {
    glEnableClientState(GL_NORMAL_ARRAY);
    glNormalPointer(GL_FLOAT, 0, (float*)normals);
  }

  if (tcs) {
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glTexCoordPointer(2, GL_FLOAT, 0, tcs);
  }

  if (!connectivity) {

    glDrawArrays(GL_POINTS, 0, vertsN);

  }
  else {

    glDrawElements(GL_TRIANGLES, 3*trianglesN, GL_UNSIGNED_INT, connectivity);

  }

  glDisableClientState(GL_VERTEX_ARRAY);

  if (colors) {
    glDisableClientState(GL_COLOR_ARRAY);
  }

  if (normals) {
    glDisableClientState(GL_NORMAL_ARRAY);
  }

  if (tcs) {
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  }

}

TMesh::TMesh(FrameBuffer *fb, PPC *ppc) {

  int w = fb->w;
  int h = fb->h;
  Initialize();
  vertsN = fb->w*fb->h;
  verts = new V3[vertsN];
  colors = new V3[vertsN];
  for (int v = 0; v < fb->h; v++) {
    for (int u = 0; u < fb->w; u++) {
      int uv = (fb->h-1-v)*fb->w+u;
      verts[uv] = ppc->Unproject(u, v, fb->zb[uv]);
      colors[uv].SetFromColor(fb->Get(u, v));
    }
  }

  connectivity = new unsigned int[(w-1)*(h-1)*2];
  trianglesN = 0;
  for (int v = 0; v < fb->h-1; v++) {
    for (int u = 0; u < fb->w-1; u++) {
      int uv = (fb->h-1-v)*fb->w+u;
      if (
        verts[uv][0] == -FLT_MAX || 
        verts[uv+1][0] == -FLT_MAX || 
        verts[uv+1-w][0] == -FLT_MAX || 
        verts[uv-w][0] == -FLT_MAX)
        continue;
      connectivity[3*trianglesN+0] = uv;
      connectivity[3*trianglesN+1] = uv-w;
      connectivity[3*trianglesN+2] = uv-w+1;
      trianglesN++;
      connectivity[3*trianglesN+0] = uv;
      connectivity[3*trianglesN+1] = uv-w+1;
      connectivity[3*trianglesN+2] = uv+1;
      trianglesN++;
    }
  }


}
